/* -----------------------------------------------------------------------------
 * Copyright (c) 2013-2014 ARM Ltd.
 *
 * This software is provided 'as-is', without any express or implied warranty.
 * In no event will the authors be held liable for any damages arising from
 * the use of this software. Permission is granted to anyone to use this
 * software for any purpose, including commercial applications, and to alter
 * it and redistribute it freely, subject to the following restrictions:
 *
 * 1. The origin of this software must not be misrepresented; you must not
 *    claim that you wrote the original software. If you use this software in
 *    a product, an acknowledgment in the product documentation would be
 *    appreciated but is not required.
 *
 * 2. Altered source versions must be plainly marked as such, and must not be
 *    misrepresented as being the original software.
 *
 * 3. This notice may not be removed or altered from any source distribution.
 *
 *
 * $Date:        20. November 2014
 * $Revision:    V2.01
 *
 * Driver:       Driver_USBD0
 * Configured:   via RTE_Device.h configuration file
 * Project:      USB Device Driver for SiliconLabs EFM32GGxxx
 * --------------------------------------------------------------------------
 * Use the following configuration settings in the middleware component
 * to connect to this driver.
 *
 *   Configuration Setting                Value
 *   ---------------------                -----
 *   Connect to hardware via Driver_USBD# = 0
 * -------------------------------------------------------------------------- */

/* History:
 *  Version 2.01
 *    Send data on first NAK to prevent simultaneous write to FIFO from library
 *    and from IRQ routine
 *  Version 2.00
 *    Updated to 2.00 API
 *  Version 1.04
 *    Multiple packet read
 *  Version 1.03
 *    Based on API V1.10 (namespace prefix ARM_ added)
 *  Version 1.00
 *    Initial release
 */

/* NOTE:
 *  - Isochronous Endpoints are not supported
 */

#include <stdint.h>
#include <string.h>
#include "cmsis_os.h"

#include "Driver_USBD.h"

#include "em_device.h"
#include "em_cmu.h"

#include "RTE_Device.h"
#include "RTE_Components.h"

extern uint8_t USB_role;

extern bool USB_PinsConfigure   (uint8_t pins_mask);
extern bool USB_PinsUnconfigure (uint8_t pins_mask);

// USBD Driver *****************************************************************

#define ARM_USBD_DRV_VERSION ARM_DRIVER_VERSION_MAJOR_MINOR(2,1) // USBD driver version

// Driver Version
static const ARM_DRIVER_VERSION usbd_driver_version = { ARM_USBD_API_VERSION, ARM_USBD_DRV_VERSION };

// Driver Capabilities
static const ARM_USBD_CAPABILITIES usbd_driver_capabilities = {
  0,  // VBUS detection
  0,  // Signal VBUS On event
  0,  // Signal VBUS Off event
};

// Number of Endpoints (excluding Endpoint 0, max = 6)
#ifndef USBD_EP_NUM
#define USBD_EP_NUM             6
#endif

#define USB_RX_FIFO_SIZE        640
#define USB_TX0_FIFO_SIZE       80
#define USB_TX1_FIFO_SIZE       80
#define USB_TX2_FIFO_SIZE       80
#define USB_TX3_FIFO_SIZE       80
#define USB_TX4_FIFO_SIZE       80
#define USB_TX5_FIFO_SIZE       80
#define USB_TX6_FIFO_SIZE       80

#define USB_TX_FIFO(n)          *((volatile uint32_t*)(USB->FIFO0D + n*1024))
#define USB_RX_FIFO             *((volatile uint32_t*)(USB->FIFO0D))

#define USB_DIEPTSIZ(EPNum)     *(&USB->DIEP0TSIZ    + EPNum * 8)
#define USB_DIEPCTL(EPNum)      *(&USB->DIEP0CTL     + EPNum * 8)
#define USB_DTXFSTS(EPNum)      *(&USB->DIEP0TXFSTS  + EPNum * 8)
#define USB_DOEPTSIZ(EPNum)     *(&USB->DOEP0TSIZ    + EPNum * 8)
#define USB_DOEPCTL(EPNum)      *(&USB->DOEP0CTL     + EPNum * 8)
#define USB_DIEPINT(EPNum)      *(&USB->DIEP0INT     + EPNum * 8)
#define USB_DOEPINT(EPNum)      *(&USB->DOEP0INT     + EPNum * 8)

#define USB_EP_IN_TYPE(num)      ((USB_DIEPCTL(num) >> 18) & 3)
#define USB_EP_OUT_TYPE(num)     ((USB_DOEPCTL(num) >> 18) & 3)

// USB Driver flags
#define USB_INITIALIZED          (1 << 0)
#define USB_POWERED              (1 << 1)
#define USB_CONNECTED            (1 << 2)

// USB Device Endpoint flags
#define USBD_EP_FLAG_CONFIGURED  (1 << 0)
#define USBD_EP_FLAG_BUSY        (1 << 1)

// Endpoint number and index
#define EP_NUM(ep_addr)           (ep_addr & ARM_USB_ENDPOINT_NUMBER_MASK)
#define EP_ID(ep_addr)            (EP_NUM(ep_addr) * 2 + ((ep_addr >> 7) & 1))

// Endpoint structure
typedef struct {
  uint8_t  *buffer;
  uint32_t  bufferIndex;
  uint32_t  dataSize;
  uint32_t  maxPacketSize;
  uint8_t   in_NAK;
  uint8_t   flags;
} ENDPOINT;

// Local variables and structures
static ARM_USBD_SignalDeviceEvent_t   cbDeviceEvent;
static ARM_USBD_SignalEndpointEvent_t cbEndpointEvent;

static uint32_t usbd_flags = 0;
static ARM_USBD_STATE usbd_state = {0};

         static uint8_t  setup_buf[8];
volatile static uint8_t  setup_flag = 0;

volatile static ENDPOINT ep[(USBD_EP_NUM + 1)*2];

// Function prototypes
static uint16_t  USBD_GetFrameNumber (void);


// Local functions

/**
  \fn          void USBD_FlushInEpFifo (uint8_t ep_addr)
  \brief       Flush IN Endpoint FIFO
  \param[in]   ep_addr specifies Endpoint Address
                 ep_addr.0..3: Address
                 ep_addr.7:    Direction
*/
static void USBD_FlushInEpFifo (uint8_t ep_addr) {
  ep_addr &= ARM_USB_ENDPOINT_NUMBER_MASK;

  // Flush transmit FIFO
  USB->GRSTCTL = (USB->GRSTCTL & ~_USB_GRSTCTL_TXFNUM_MASK) |
                 (ep_addr << _USB_GRSTCTL_TXFNUM_SHIFT)     |
                  USB_GRSTCTL_TXFFLSH;
}

/**
  \fn          void USBD_Reset (void)
  \brief       Called after usbd reset interrupt to reset configuration
*/
static void USBD_Reset (void) {
  uint32_t i;

  // Clear Endpoint mask registers
  USB->DOEPMSK = 0;
  USB->DIEPMSK = 0;

  for (i = 0; i <= USBD_EP_NUM; i++) {
    if (USB_DOEPCTL(i) & USB_DOEP_CTL_EPENA) {
      USB_DOEPCTL(i)   = USB_DOEP_CTL_EPDIS |       // Endpoint disable
                         USB_DOEP_CTL_SNAK;         // Endpoint set NAK
    }
    if (USB_DIEPCTL(i) & USB_DIEP_CTL_EPENA) {
      USB_DIEPCTL(i)   = USB_DIEP_CTL_EPDIS |       // Endpoint disable
                         USB_DIEP_CTL_SNAK;         // Endpoint set NAK
    }
    USBD_FlushInEpFifo(i);

    // Reset Endpoint resources
    memset((void *)(&ep[i*2  ]), 0, sizeof (ENDPOINT));
    memset((void *)(&ep[i*2+1]), 0, sizeof (ENDPOINT));

    // Clear IN Endpoint interrupts
    USB_DIEPINT(i) = USB_DIEP_INT_XFERCOMPL   |
                     USB_DIEP_INT_EPDISBLD    |
                     USB_DIEP_INT_INTKNTXFEMP |
                     USB_DIEP_INT_INEPNAKEFF  |
                     USB_DIEP_INT_TXFEMP      ;

    // Clear OUT Endpoint interrupts
    USB_DOEPINT(i) = USB_DOEP_INT_XFERCOMPL   |
                     USB_DOEP_INT_EPDISBLD    |
                     USB_DOEP_INT_SETUP       |
                     USB_DOEP_INT_OUTTKNEPDIS ;
  }

  // Set device address to 0
  USB->DCFG       = (USB->DCFG & ~_USB_DCFG_DEVADDR_MASK) | (0 << _USB_DCFG_DEVADDR_SHIFT);
  USB->DAINTMSK   =  USB_DAINTMSK_INEPMSK0 |        // Enable IN Endpoint0 interrupt
                     USB_DAINTMSK_OUTEPMSK0;        // Enable OUT Endpoint0 interrupt

  // Enable Setup phase done and OUT transfer complete interrupt
  USB->DOEPMSK    = USB_DOEPMSK_SETUPMSK      |
                    USB_DOEPMSK_XFERCOMPLMSK;

  // Enable IN transfer complete interrupt
  USB->DIEPMSK    = USB_DIEPMSK_XFERCOMPLMSK;

  // Configure FIFOs
  USB->GRXFSIZ   =  USB_RX_FIFO_SIZE   / 4;
  USB->GNPTXFSIZ = (USB_RX_FIFO_SIZE   / 4) |
                   ((USB_TX0_FIFO_SIZE / 4) << 16);

  USB->DIEPTXF1  = ((USB_RX_FIFO_SIZE + USB_TX0_FIFO_SIZE) / 4) | ((USB_TX1_FIFO_SIZE / 4) << 16);
  USB->DIEPTXF2  = ((USB_RX_FIFO_SIZE + USB_TX0_FIFO_SIZE + USB_TX1_FIFO_SIZE) / 4) | ((USB_TX2_FIFO_SIZE / 4) << 16);
  USB->DIEPTXF3  = ((USB_RX_FIFO_SIZE + USB_TX0_FIFO_SIZE + USB_TX1_FIFO_SIZE + USB_TX2_FIFO_SIZE) / 4) | ((USB_TX3_FIFO_SIZE / 4) << 16);
  USB->DIEPTXF4  = ((USB_RX_FIFO_SIZE + USB_TX0_FIFO_SIZE + USB_TX1_FIFO_SIZE + USB_TX2_FIFO_SIZE + USB_TX3_FIFO_SIZE) / 4) | ((USB_TX4_FIFO_SIZE / 4) << 16);
  USB->DIEPTXF5  = ((USB_RX_FIFO_SIZE + USB_TX0_FIFO_SIZE + USB_TX1_FIFO_SIZE + USB_TX2_FIFO_SIZE + USB_TX3_FIFO_SIZE + USB_TX4_FIFO_SIZE) / 4) | ((USB_TX5_FIFO_SIZE / 4) << 16);
  USB->DIEPTXF6  = ((USB_RX_FIFO_SIZE + USB_TX0_FIFO_SIZE + USB_TX1_FIFO_SIZE + USB_TX2_FIFO_SIZE + USB_TX3_FIFO_SIZE + USB_TX4_FIFO_SIZE + USB_TX5_FIFO_SIZE) / 4) | ((USB_TX6_FIFO_SIZE / 4) << 16);
}

/**
  \fn          void USBD_EndpointReadSet (uint8_t ep_addr)
  \brief       Set Endpoint for next read.
  \param[in]   ep_addr specifies Endpoint Address
                 ep_addr.0..3: Address
                 ep_addr.7:    Direction
*/
static void USBD_EndpointReadSet (uint8_t ep_addr) {
  volatile ENDPOINT *ptr_ep;
  uint32_t           sz;
  uint8_t            ep_num;

  ptr_ep = &ep[EP_ID(ep_addr)];
  ep_num = EP_NUM(ep_addr);

  // Set packet count and transfer size
  if  (ptr_ep->dataSize > ptr_ep->maxPacketSize) sz = ptr_ep->maxPacketSize;
  else                                           sz = ptr_ep->dataSize;

  if (ep_num) {
    USB_DOEPTSIZ(ep_num) = (1 << _USB_DOEP_TSIZ_PKTCNT_SHIFT) |
                            sz;
  } else {
    USB_DOEPTSIZ(0)      = (1 << _USB_DOEP0TSIZ_PKTCNT_SHIFT) |
                           (3 << _USB_DOEP0TSIZ_SUPCNT_SHIFT) |
                            sz;
  }

  // Clear NAK and enable Endpoint
  USB_DOEPCTL(ep_num) |= USB_DOEP_CTL_EPENA | USB_DOEP_CTL_CNAK;
}

/**
  \fn          int32_t USBD_ReadFromFifo (uint8_t ep_addr, uint32_t sz)
  \brief       Read data from USB Endpoint.
  \param[in]   ep_addr specifies Endpoint Address
                 ep_addr.0..3: Address
                 ep_addr.7:    Direction
  \param[in]   sz specifies data size to be read from FIFO
  \return      number of data bytes read
*/
static int32_t USBD_ReadFromFifo (uint8_t ep_addr, uint32_t sz) {
  volatile ENDPOINT *ptr_ep;
  uint32_t           i, residue;
  uint8_t           *ptr_dest_8;
  __packed uint32_t *ptr_dest_32;
  volatile uint32_t *ptr_src;
  uint8_t            ep_num;
  uint8_t            tmp_buf[4];

  ptr_ep = &ep[EP_ID(ep_addr)];
  ep_num = EP_NUM(ep_addr);

  // Check if Endpoint is activated and buffer available
  if (!(USB_DOEPCTL(ep_num) & USB_DOEP_CTL_USBACTEP)) return 0;
  if (!ptr_ep->buffer)                                return 0;

  if (sz > ptr_ep->dataSize)
    sz = ptr_ep->dataSize;

  // Copy data from FIFO
  ptr_src     = (volatile uint32_t *)(USB->FIFO0D);
  ptr_dest_32 = (__packed uint32_t *)(ptr_ep->buffer + ptr_ep->bufferIndex);
  i           = sz / 4;
  while (i--) {
    *ptr_dest_32++ = *ptr_src;
  }
  ptr_ep->bufferIndex += (sz / 4) * 4;

  // If data size is not equal n*4
  residue = sz % 4;
  if (residue) {
    ptr_dest_8 = (uint8_t *)(ptr_dest_32);
    *((__packed uint32_t *)tmp_buf) = USB_RX_FIFO;
    for (i = 0; i < residue; i++) {
      *ptr_dest_8++ = tmp_buf[i];
      ptr_ep->bufferIndex ++;
    }
  }

  if (sz != ptr_ep->maxPacketSize) ptr_ep->dataSize  = 0;
  else                             ptr_ep->dataSize -= sz;

  return sz;
}

/**
  \fn          void USBD_WriteToFifo (uint8_t ep_addr)
  \brief       Write data to Endpoint FIFO.
  \param[in]   ep_addr specifies Endpoint Address
                 ep_addr.0..3: Address
                 ep_addr.7:    Direction
*/
static void USBD_WriteToFifo (uint8_t ep_addr) {
  volatile ENDPOINT *ptr_ep;
  uint8_t            ep_num;
  uint32_t           sz, i;
  volatile uint32_t *ptr_dest;
  __packed uint32_t *ptr_src;

  ptr_ep = &ep[EP_ID(ep_addr)];
  ep_num = EP_NUM(ep_addr);

  if (ptr_ep->dataSize > ptr_ep->maxPacketSize) sz = ptr_ep->maxPacketSize;
  else                                          sz = ptr_ep->dataSize;

  // Check if enough space in FIFO
  if ((USB_DTXFSTS(ep_num) * 4) < sz) return;

  // Set transfer size and packet count
  USB_DIEPTSIZ(ep_num) = (1 << _USB_DIEP_TSIZ_PKTCNT_SHIFT) |
                         (1 << _USB_DIEP_TSIZ_MC_SHIFT)     |
                          sz;

  // Enable Endpoint and clear NAK
  USB_DIEPCTL(ep_num) |= USB_DIEP_CTL_EPENA | USB_DIEP_CTL_CNAK;

  ptr_src  = (__packed uint32_t *)(ptr_ep->buffer + ptr_ep->bufferIndex);
  ptr_dest = (volatile uint32_t *)(USB->FIFO0D + ep_num*1024);

  ptr_ep->bufferIndex += sz;
  ptr_ep->dataSize    -= sz;
  i = (sz+3)>>2;
  // Copy data to FIFO
  while (i--) {
    *ptr_dest = *ptr_src++;
  }
}


// USB DEVICE DRIVER FUNCTIONS

/**
  \fn          ARM_DRIVER_VERSION USBD_GetVersion (void)
  \brief       Get driver version.
  \return      \ref ARM_DRIVER_VERSION
*/
static ARM_DRIVER_VERSION USBD_GetVersion (void) { return usbd_driver_version; }

/**
  \fn          ARM_USBD_CAPABILITIES USBD_GetCapabilities (void)
  \brief       Get driver capabilities.
  \return      \ref ARM_USBD_CAPABILITIES
*/
static ARM_USBD_CAPABILITIES USBD_GetCapabilities (void) { return usbd_driver_capabilities; }

/**
  \fn          int32_t USBD_Initialize (ARM_USBD_SignalDeviceEvent_t   cb_device_event,
                                        ARM_USBD_SignalEndpointEvent_t cb_endpoint_event)
  \brief       Initialize USB Device Interface.
  \param[in]   cb_device_event    Pointer to \ref ARM_USBD_SignalDeviceEvent
  \param[in]   cb_endpoint_event  Pointer to \ref ARM_USBD_SignalEndpointEvent
  \return      \ref execution_status
*/
static int32_t USBD_Initialize (ARM_USBD_SignalDeviceEvent_t   cb_device_event,
                                ARM_USBD_SignalEndpointEvent_t cb_endpoint_event) {

  if (usbd_flags & USB_INITIALIZED) return ARM_DRIVER_OK;
  if (usbd_flags)                   return ARM_DRIVER_ERROR;

  cbDeviceEvent   = cb_device_event;
  cbEndpointEvent = cb_endpoint_event;

  USB_role        = ARM_USB_ROLE_DEVICE;
  USB_PinsConfigure (ARM_USB_PIN_DP | ARM_USB_PIN_DM);

  usbd_flags      = USB_INITIALIZED;

  return ARM_DRIVER_OK;
}

/**
  \fn          int32_t USBD_Uninitialize (void)
  \brief       De-initialize USB Device Interface.
  \return      \ref execution_status
*/
static int32_t USBD_Uninitialize (void) {

  if (!(usbd_flags & USB_INITIALIZED)) return ARM_DRIVER_OK;
  if (  usbd_flags & USB_POWERED     ) return ARM_DRIVER_ERROR;

  USB_PinsUnconfigure (ARM_USB_PIN_DP | ARM_USB_PIN_DM);
  USB_role   = ARM_USB_ROLE_NONE;

  usbd_flags = 0;

  return ARM_DRIVER_OK;
}

/**
  \fn          int32_t USBD_PowerControl (ARM_POWER_STATE state)
  \brief       Control USB Device Interface Power.
  \param[in]   state  Power state
  \return      \ref execution_status
*/
static int32_t USBD_PowerControl (ARM_POWER_STATE state) {

  if (!(usbd_flags & USB_INITIALIZED)) return ARM_DRIVER_ERROR;

  switch (state) {
    case ARM_POWER_OFF:
      if (!(usbd_flags & USB_POWERED)) return ARM_DRIVER_OK;
      usbd_flags   &= ~USB_POWERED;
      NVIC_DisableIRQ(USB_IRQn);                        // Disable interrupt
      USB->GAHBCFG &= ~USB_GAHBCFG_GLBLINTRMSK;         // Disable global interrupt mask

      USB->DCTL    |=  USB_DCTL_SFTDISCON;              // Soft disconnect enabled

      // Disable USB clock
      CMU->HFCORECLKEN0 &= ~(CMU_HFCORECLKEN0_USB | CMU_HFCORECLKEN0_USBC);
      break;

    case ARM_POWER_FULL:
      if (usbd_flags & USB_POWERED) return ARM_DRIVER_OK;
      CMU->HFCORECLKEN0 |= CMU_HFCORECLKEN0_USBC;
      CMU_ClockSelectSet( cmuClock_USBC, cmuSelect_HFCLK );
      osDelay(100);
      CMU->HFCORECLKEN0 |=  CMU_HFCORECLKEN0_USB;       // Enable USB clock

      // Wait until AHB Master state machine is in the idle condition
      while (!(USB->GRSTCTL & USB_GRSTCTL_AHBIDLE));

      // Core soft reset
      USB->GRSTCTL      |=  USB_GRSTCTL_CSFTRST;
      while (USB->GRSTCTL & USB_GRSTCTL_CSFTRST);

      osDelay(10);

      USB->GAHBCFG      &= ~USB_GAHBCFG_GLBLINTRMSK;    // Disable global interrupt mask
      USB->DCTL         |=  USB_DCTL_SFTDISCON;         // Soft disconnect enabled

      // Set turnaround time and force device mode
      USB->GUSBCFG   =  (USB->GUSBCFG & ~_USB_GUSBCFG_USBTRDTIM_MASK) |
                        (5 << _USB_GUSBCFG_USBTRDTIM_SHIFT)           |
                         USB_GUSBCFG_FORCEDEVMODE;

      osDelay(100);

      // Clear interrupts
      USB->GINTSTS   =   USB_GINTSTS_CURMOD       |
                         USB_GINTSTS_SOF          |
                         USB_GINTSTS_ERLYSUSP     |
                         USB_GINTSTS_USBSUSP      |
                         USB_GINTSTS_USBRST       |
                         USB_GINTSTS_ENUMDONE     |
                         USB_GINTSTS_ISOOUTDROP   |
                         USB_GINTSTS_INCOMPISOIN  |
                         USB_GINTSTS_INCOMPLP     |
                         USB_GINTSTS_FETSUSP      |
                         USB_GINTSTS_RESETDET     |
                         USB_GINTSTS_CONIDSTSCHNG |
                         USB_GINTSTS_DISCONNINT   |
                         USB_GINTSTS_SESSREQINT   |
                         USB_GINTSTS_WKUPINT;

      // Unmask interrupts
      USB->GINTMSK   =   USB_GINTSTS_USBSUSP      |
                         USB_GINTSTS_USBRST       |
                         USB_GINTSTS_ENUMDONE     |
                         USB_GINTSTS_RXFLVL       |
                         USB_GINTSTS_IEPINT       |
                         USB_GINTSTS_OEPINT       |
                         USB_GINTSTS_WKUPINT;

      usbd_flags    |= USB_POWERED;

      NVIC_EnableIRQ(USB_IRQn);                         // Enable interrupts
      USB->GAHBCFG  |=  USB_GAHBCFG_GLBLINTRMSK |
                        USB_GAHBCFG_NPTXFEMPLVL ;
      break;

    default:
      return ARM_DRIVER_ERROR_UNSUPPORTED;
  }

  return ARM_DRIVER_OK;
}

/**
  \fn          int32_t USBD_DeviceConnect (void)
  \brief       Connect USB Device.
  \return      \ref execution_status
*/
static int32_t USBD_DeviceConnect (void) {

  if (!(usbd_flags & USB_POWERED) ) return ARM_DRIVER_ERROR;
  if (  usbd_flags & USB_CONNECTED) return ARM_DRIVER_OK;

  USB->DCTL  &= ~USB_DCTL_SFTDISCON;    // Soft disconnect disabled

  usbd_flags |=  USB_CONNECTED;

  return ARM_DRIVER_OK;
}

/**
  \fn          int32_t USBD_DeviceDisconnect (void)
  \brief       Disconnect USB Device.
  \return      \ref execution_status
*/
static int32_t USBD_DeviceDisconnect (void) {

  if (!(usbd_flags & USB_POWERED)  ) return ARM_DRIVER_ERROR;
  if (!(usbd_flags & USB_CONNECTED)) return ARM_DRIVER_OK;

  USB->DCTL  |=  USB_DCTL_SFTDISCON;    // Soft disconnect enabled

  usbd_state.active = false;

  usbd_flags &= ~USB_CONNECTED;

  return ARM_DRIVER_OK;
}

/**
  \fn          ARM_USBD_STATE USBD_DeviceGetState (void)
  \brief       Get current USB Device State.
  \return      Device State \ref ARM_USBD_STATE
*/
static ARM_USBD_STATE USBD_DeviceGetState (void) {
  return usbd_state;
}

/**
  \fn          int32_t USBD_DeviceRemoteWakeup (void)
  \brief       Trigger USB Remote Wakeup.
  \return      \ref execution_status
*/
static int32_t USBD_DeviceRemoteWakeup (void) {

  if (!(usbd_flags & USB_CONNECTED)) return ARM_DRIVER_ERROR;

  USB->DCTL |=   USB_DCTL_RMTWKUPSIG;   // Remote wakeup signalling
  osDelay(5);
  USB->DCTL &=  ~USB_DCTL_RMTWKUPSIG;

  return ARM_DRIVER_OK;
}

/**
  \fn          int32_t USBD_DeviceSetAddress (uint8_t dev_addr)
  \brief       Set USB Device Address.
  \param[in]   dev_addr  Device Address
  \return      \ref execution_status
*/
static int32_t USBD_DeviceSetAddress (uint8_t dev_addr) {

  if (!(usbd_flags & USB_CONNECTED)) return ARM_DRIVER_ERROR;

  USB->DCFG = (USB->DCFG & ~_USB_DCFG_DEVADDR_MASK) |
              (dev_addr << _USB_DCFG_DEVADDR_SHIFT) ;

  return ARM_DRIVER_OK;
}

/**
  \fn          int32_t USBD_ReadSetupPacket (uint8_t *setup)
  \brief       Read setup packet received over Control Endpoint.
  \param[out]  setup  Pointer to buffer for setup packet
  \return      \ref execution_status
*/
static int32_t USBD_ReadSetupPacket (uint8_t *setup) {

  if (!(usbd_flags & USB_CONNECTED)) return ARM_DRIVER_ERROR;
  if (!setup_flag)                   return ARM_DRIVER_ERROR;

  setup_flag = 0;
  memcpy(setup, setup_buf, 8);

  if (setup_flag) {
    // New setup packet was received while this was being read
    return ARM_DRIVER_ERROR;
  }

  return ARM_DRIVER_OK;
}

/**
  \fn          int32_t USBD_EndpointConfigure (uint8_t  ep_addr,
                                               uint8_t  ep_type,
                                               uint16_t ep_max_packet_size)
  \brief       Configure USB Endpoint.
  \param[in]   ep_addr  Endpoint Address
                - ep_addr.0..3: Address
                - ep_addr.7:    Direction
  \param[in]   ep_type  Endpoint Type (ARM_USB_ENDPOINT_xxx)
  \param[in]   ep_max_packet_size Endpoint Maximum Packet Size
  \return      \ref execution_status
*/
static int32_t USBD_EndpointConfigure (uint8_t  ep_addr,
                                       uint8_t  ep_type,
                                       uint16_t ep_max_packet_size) {
  volatile ENDPOINT *ptr_ep;
  uint32_t           ep_num, ep_dir, ep_mps;

  ptr_ep = &ep[EP_ID(ep_addr)];

  if (!(usbd_flags & USB_CONNECTED))           return ARM_DRIVER_ERROR;
  if (ptr_ep->flags & USBD_EP_FLAG_BUSY)       return ARM_DRIVER_ERROR_BUSY;
  if (ptr_ep->flags & USBD_EP_FLAG_CONFIGURED) return ARM_DRIVER_OK;

  ep_num =  EP_NUM(ep_addr);
  ep_dir = (ep_addr & ARM_USB_ENDPOINT_DIRECTION_MASK) == ARM_USB_ENDPOINT_DIRECTION_MASK;
  ep_mps =  ep_max_packet_size & ARM_USB_ENDPOINT_MAX_PACKET_SIZE_MASK;

  // Check if Endpoint is enabled
  if (ep_num > USBD_EP_NUM) return ARM_DRIVER_ERROR;

  // Set Endpoint resources
  ptr_ep->buffer        = NULL;
  ptr_ep->dataSize      = 0;
  ptr_ep->maxPacketSize = ep_mps;

  if (ep_dir) {                         // IN Endpoint
    ptr_ep->in_NAK = 0;                                 // Clear IN Endpoint NAK flag

    // Configure IN Endpoint
    USB_DIEPCTL(ep_num) = (ep_num  << _USB_DIEP_CTL_TXFNUM_SHIFT) |     // FIFO Number
                          (ep_type << _USB_DIEP_CTL_EPTYPE_SHIFT) |     // Endpoint Type
                           ep_mps;                                      // Max Packet Size

    // Set DATA0 PID for Interrupt or Bulk Endpoint
    if (ep_type >= ARM_USB_ENDPOINT_BULK)
      USB_DIEPCTL(ep_num) |= USB_DIEP_CTL_SETD0PIDEF;

    USB_DIEPCTL(ep_num)   |= USB_DIEP_CTL_USBACTEP;     // Activate Endpoint

    if (USB_DIEPCTL(ep_num) & USB_DIEP_CTL_EPENA)
      USB_DIEPCTL(ep_num) |= USB_DIEP_CTL_EPDIS;        // Disable Endpoint

    USBD_FlushInEpFifo (ep_addr);

  } else {                              // OUT Endpoint

    // Configure OUT Endpoint
    USB_DOEPCTL(ep_num) = (ep_type <<  _USB_DOEP_CTL_EPTYPE_SHIFT) |    // Endpoint Type
                           USB_DOEP_CTL_SNAK                       |    // Set NAK
                           ep_mps;                                      // Max Packet Size

    // Set DATA0 PID for Interrupt or Bulk Endpoint
    if (ep_type >= ARM_USB_ENDPOINT_BULK)
      USB_DOEPCTL(ep_num) |= USB_DOEP_CTL_SETD0PIDEF;

    USB_DOEPCTL(ep_num) |= USB_DOEP_CTL_USBACTEP;       // Activate Endpoint
  }

  ptr_ep->flags |= USBD_EP_FLAG_CONFIGURED;             // Set Endpoint configured flag

  USB->DAINTMSK |= (1 << (ep_num + ((ep_dir ^ 1) << 4)));               // Enable Endpoint interrupt

  return ARM_DRIVER_OK;
}

/**
  \fn          int32_t USBD_EndpointUnconfigure (uint8_t ep_addr)
  \brief       Unconfigure USB Endpoint.
  \param[in]   ep_addr  Endpoint Address
                - ep_addr.0..3: Address
                - ep_addr.7:    Direction
  \return      \ref execution_status
*/
static int32_t USBD_EndpointUnconfigure (uint8_t ep_addr) {
  volatile ENDPOINT *ptr_ep;
  uint32_t           ep_num, ep_dir;

  ptr_ep = &ep[EP_ID(ep_addr)];

  if (!(usbd_flags & USB_CONNECTED))              return ARM_DRIVER_ERROR;
  if (  ptr_ep->flags & USBD_EP_FLAG_BUSY)        return ARM_DRIVER_ERROR_BUSY;
  if (!(ptr_ep->flags & USBD_EP_FLAG_CONFIGURED)) return ARM_DRIVER_OK;

  ep_num =  EP_NUM(ep_addr);
  ep_dir = (ep_addr & ARM_USB_ENDPOINT_DIRECTION_MASK) == ARM_USB_ENDPOINT_DIRECTION_MASK;

  USB->DAINTMSK &= ~(1 << (ep_num + ((ep_dir ^ 1) << 4)));              // Disable Endpoint interrupt

  ptr_ep->flags &= ~USBD_EP_FLAG_CONFIGURED;            // Clear Endpoint configured flag

  memset((void *)(ptr_ep), 0, sizeof (ENDPOINT));       // Reset Endpoint parameters

  if (ep_dir) {                         // IN Endpoint

    if (USB_DIEPCTL(ep_num) & USB_DIEP_CTL_EPENA)
      USB_DIEPCTL(ep_num)  |=  USB_DIEP_CTL_EPDIS;      // Disable Endpoint

    USB_DIEPCTL(ep_num)    |=  USB_DIEP_CTL_SNAK;       // Set Endpoint NAK

    if (ep_num) USB_DIEPCTL(ep_num) &= ~USB_DIEP_CTL_USBACTEP;          // Deactivate Endpoint
  } else {                              // OUT Endpoint
    USB->DCTL |= USB_DCTL_SGOUTNAK;                     // Set Global OUT NAK
    while (!(USB->GINTSTS & USB_GINTSTS_GOUTNAKEFF));

    USB_DOEPCTL(ep_num) |= USB_DOEP_CTL_SNAK;           // Set Endpoint NAK

    if (ep_num) {
      // Disable OUT Endpoint
      if (USB_DOEPCTL(ep_num) & USB_DOEP_CTL_EPENA) {   // If Endpoint is Enabled
        USB_DOEPCTL(ep_num)  |=  USB_DOEP_CTL_EPDIS;    // Disable Endpoint
        while (!(USB_DOEPINT(ep_num) & USB_DOEP_INT_EPDISBLD));
      }
      USB_DOEPCTL(ep_num) &= ~USB_DOEP_CTL_USBACTEP;    // Deactivate Endpoint
    }

    USB->DCTL |= USB_DCTL_CGOUTNAK;                     // Clear Global OUT NAK
  }

  return ARM_DRIVER_OK;
}

/**
  \fn          int32_t USBD_EndpointStall (uint8_t ep_addr, bool stall)
  \brief       Set/Clear Stall for USB Endpoint.
  \param[in]   ep_addr  Endpoint Address
                - ep_addr.0..3: Address
                - ep_addr.7:    Direction
  \param[in]   stall  Operation
                - \b false Clear
                - \b true Set
  \return      \ref execution_status
*/
static int32_t USBD_EndpointStall (uint8_t ep_addr, bool stall) {
  volatile ENDPOINT *ptr_ep;
  uint32_t           ep_num, ep_dir;

  ptr_ep = &ep[EP_ID(ep_addr)];

  if (!(usbd_flags & USB_CONNECTED))              return ARM_DRIVER_ERROR;
  if (!(ptr_ep->flags & USBD_EP_FLAG_CONFIGURED)) return ARM_DRIVER_ERROR;
  if (  ptr_ep->flags & USBD_EP_FLAG_BUSY       ) return ARM_DRIVER_ERROR_BUSY;

  ep_num =  EP_NUM(ep_addr);
  ep_dir = (ep_addr & ARM_USB_ENDPOINT_DIRECTION_MASK) == ARM_USB_ENDPOINT_DIRECTION_MASK;

  if (stall) {                          // Activate STALL
    ptr_ep->buffer      = NULL;
    ptr_ep->dataSize    = 0;
    ptr_ep->bufferIndex = 0;

    if (ep_dir) {                       // IN Endpoint
      if (USB_DIEPCTL(ep_num) & USB_DIEP_CTL_EPENA)
        USB_DIEPCTL(ep_num)  |= USB_DIEP_CTL_STALL | USB_DIEP_CTL_EPDIS;
      else
        USB_DIEPCTL(ep_num)  |= USB_DIEP_CTL_STALL;

      USBD_FlushInEpFifo (ep_addr);
    } else {                            // OUT Endpoint
      USB->DCTL |= USB_DCTL_SGOUTNAK;                   // Set Global OUT NAK
      while (!(USB->GINTSTS & USB_GINTSTS_GOUTNAKEFF));

      // Stall OUT Endpoint
      if (USB_DOEPCTL(ep_num) & USB_DOEP_CTL_EPENA)
        USB_DOEPCTL(ep_num)  |= USB_DOEP_CTL_STALL | USB_DOEP_CTL_EPDIS;
      else
        USB_DOEPCTL(ep_num)  |= USB_DOEP_CTL_STALL;

      USB->DCTL |= USB_DCTL_CGOUTNAK;                   // Clear global NAK
    }
  } else {                              // Clear STALL
    if (ep_dir) {                       // IN Endpoint
      if (USB_DIEPCTL(ep_num) &  USB_DIEP_CTL_EPENA)    // If Endpoint enabled
        USB_DIEPCTL(ep_num)   |= USB_DIEP_CTL_EPDIS;    // Disable Endpoint

      USBD_FlushInEpFifo (ep_addr);

      // Set DATA0 PID for Interrupt and Bulk Endpoint
      if (((USB_DIEPCTL(ep_num) & _USB_DIEP_CTL_EPTYPE_MASK) >> _USB_DIEP_CTL_EPTYPE_SHIFT) > 1)
        USB_DIEPCTL(ep_num) |= USB_DIEP_CTL_SETD0PIDEF;

      USB_DIEPCTL(ep_num) &= ~USB_DIEP_CTL_STALL;       // Clear Stall
    } else {                            // Clear OUT Endpoint stall
      // Set DATA0 pid for Interrupt and Bulk Endpoint
      if (((USB_DOEPCTL(ep_num) & _USB_DOEP_CTL_EPTYPE_MASK) >> _USB_DOEP_CTL_EPTYPE_SHIFT) > 1) {
        USB_DOEPCTL(ep_num) |= USB_DOEP_CTL_SETD0PIDEF;
      }
      USB_DOEPCTL(ep_num) &= ~USB_DOEP_CTL_STALL;       // Clear Stall
    }
  }

  return ARM_DRIVER_OK;
}

/**
  \fn          int32_t USBD_EndpointTransfer (uint8_t ep_addr, uint8_t *data, uint32_t num)
  \brief       Read data from or Write data to USB Endpoint.
  \param[in]   ep_addr  Endpoint Address
                - ep_addr.0..3: Address
                - ep_addr.7:    Direction
  \param[out]  data Pointer to buffer for data to read or with data to write
  \param[in]   num  Number of data bytes to transfer
  \return      \ref execution_status
*/
static int32_t USBD_EndpointTransfer (uint8_t ep_addr, uint8_t *data, uint32_t num) {
  volatile ENDPOINT *ptr_ep;
  uint32_t           ep_num, ep_dir;

  ptr_ep = &ep[EP_ID(ep_addr)];

  if (!(usbd_flags & USB_CONNECTED))              return ARM_DRIVER_ERROR;
  if (!(ptr_ep->flags & USBD_EP_FLAG_CONFIGURED)) return ARM_DRIVER_ERROR;
  if (  ptr_ep->flags & USBD_EP_FLAG_BUSY       ) return ARM_DRIVER_ERROR_BUSY;

  ep_dir = (ep_addr & ARM_USB_ENDPOINT_DIRECTION_MASK) == ARM_USB_ENDPOINT_DIRECTION_MASK;

  ptr_ep->flags      |= USBD_EP_FLAG_BUSY;              // Set Endpoint busy

  ptr_ep->bufferIndex = 0;
  ptr_ep->buffer      = data;
  ptr_ep->dataSize    = num;

  if (ep_dir) {                         // IN Endpoint
    ep_num = EP_NUM(ep_addr);

    ptr_ep->in_NAK       = 1;                           // Set IN Endpoint NAK flag
    USB_DIEPCTL(ep_num) |=  USB_DIEP_CTL_CNAK;          // Clear NAK
    USB_DIEPCTL(ep_num) |=  USB_DIEP_CTL_SNAK;          // Set NAK
    USB->DIEPMSK        |=  USB_DIEPMSK_INEPNAKEFFMSK;  // Enable NAK effective interrupt
  } else {                              // OUT Endpoint
    USBD_EndpointReadSet(ep_addr);
  }

  return ARM_DRIVER_OK;
}

/**
  \fn          uint32_t USBD_EndpointTransferGetResult (uint8_t ep_addr)
  \brief       Get result of USB Endpoint transfer.
  \param[in]   ep_addr  Endpoint Address
                - ep_addr.0..3: Address
                - ep_addr.7:    Direction
  \return      number of successfully transfered data bytes
*/
static uint32_t USBD_EndpointTransferGetResult (uint8_t ep_addr) {
  return ep[EP_ID(ep_addr)].bufferIndex;
}

/**
  \fn          int32_t USBD_EndpointTransferAbort (uint8_t ep_addr)
  \brief       Abort current USB Endpoint transfer.
  \param[in]   ep_addr  Endpoint Address
                - ep_addr.0..3: Address
                - ep_addr.7:    Direction
  \return      \ref execution_status
*/
static int32_t USBD_EndpointTransferAbort (uint8_t ep_addr) {
  volatile ENDPOINT *ptr_ep;
  uint32_t           ep_num;

  ptr_ep = &ep[EP_ID(ep_addr)];

  if (!(usbd_flags & USB_CONNECTED))              return ARM_DRIVER_ERROR;
  if (!(ptr_ep->flags & USBD_EP_FLAG_CONFIGURED)) return ARM_DRIVER_ERROR;

  ep_num = EP_NUM(ep_addr);

  ptr_ep->buffer      = NULL;
  ptr_ep->dataSize    = 0;
  ptr_ep->bufferIndex = 0;

  if (ep_addr & ARM_USB_ENDPOINT_DIRECTION_MASK) {
    if (USB_DIEPCTL(ep_num) &  USB_DIEP_CTL_EPENA)      // If Endpoint enabled
      USB_DIEPCTL(ep_num)   |= USB_DIEP_CTL_EPDIS;      // Disable Endpoint

    USB_DIEPCTL(ep_num)     |= USB_DIEP_CTL_SNAK;       // Set NAK

    USBD_FlushInEpFifo (ep_addr);
  } else {
    if (USB_DOEPCTL(ep_num) &  USB_DOEP_CTL_EPENA)      // If Endpoint enabled
      USB_DOEPCTL(ep_num)   |= USB_DOEP_CTL_EPDIS;      // Disable Endpoint

    USB_DOEPCTL(ep_num)     |= USB_DOEP_CTL_SNAK;       // Set NAK
  }

  ptr_ep->flags &= ~USBD_EP_FLAG_BUSY;                  // Clear Endpoint busy

  return ARM_DRIVER_OK;
}

/**
  \fn          uint16_t USBD_GetFrameNumber (void)
  \brief       Get current USB Frame Number.
  \return      Frame Number
*/
static uint16_t USBD_GetFrameNumber (void) {
  return ((USB->DSTS & _USB_DSTS_SOFFN_MASK) >> _USB_DSTS_SOFFN_SHIFT);
}

/**
  \fn          void USBD_IRQ (uint32_t gintsts)
  \brief       USB Device Interrupt Routine (IRQ).
*/
void USBD_IRQ (uint32_t gintsts) {
  volatile ENDPOINT *ptr_ep, *ptr_ep_in;
  uint32_t           val, num, msk, sz, ep_int, i;

  // Reset interrupt
  if (gintsts & USB_GINTSTS_USBRST) {
    USBD_Reset();
    cbDeviceEvent(ARM_USBD_EVENT_RESET);
    USB->GINTSTS =  USB_GINTSTS_USBRST;
  }

  // Suspend interrupt
  if (gintsts & USB_GINTSTS_USBSUSP) {
    usbd_state.active = false;
    cbDeviceEvent(ARM_USBD_EVENT_SUSPEND);
    USB->PCGCCTL |=  USB_PCGCCTL_STOPPCLK;
    USB->GINTSTS  =  USB_GINTSTS_USBSUSP;
  }

  // Resume interrupt
  if (gintsts & USB_GINTSTS_WKUPINT) {
    usbd_state.active = true;
    cbDeviceEvent(ARM_USBD_EVENT_RESUME);
    USB->PCGCCTL &= ~USB_PCGCCTL_STOPPCLK;
    USB->GINTSTS  =  USB_GINTSTS_WKUPINT;
  }

  // Speed enumeration completed
  if (gintsts & USB_GINTSTS_ENUMDONE) {
    usbd_state.speed  = ARM_USB_SPEED_FULL;
    usbd_state.active = true;

    USB->DCTL    |= USB_DCTL_CGNPINNAK;     // Clear global IN NAK
    USB->DCTL    |= USB_DCTL_CGOUTNAK ;     // Clear global OUT NAK
    USB->GINTSTS  = USB_GINTSTS_ENUMDONE;
  }

  if (gintsts & USB_GINTSTS_RXFLVL) {
    USB->GINTMSK &= ~USB_GINTSTS_RXFLVL;

    val =  USB->GRXSTSP;
    num =  val & 0x0F;
    sz  = (val >> 4) & 0x7FF;

    switch ((val >> 17) & 0x0F) {
      // Setup packet
      case 6:
        // Read setup packet
        *(__packed uint32_t *)(setup_buf)     = USB_RX_FIFO;
        *(__packed uint32_t *)(setup_buf + 4) = USB_RX_FIFO;

        // Analyze Setup packet for SetAddress
        if (setup_buf[0] == 0) {
          if (setup_buf[1] == 5)
            USBD_DeviceSetAddress(setup_buf[2]);
        }
        setup_flag = 1;
      break;

      // OUT packet
      case 2:
        USBD_ReadFromFifo(num, sz);
        break;

      // Global OUT NAK
      case 1:
        break;

      // OUT transfer completed
      case 3:
        break;

      // SETUP transaction completed
      case 4:
        break;

      default:
        break;
    }
    USB->GINTMSK |= USB_GINTSTS_RXFLVL;
  }

  // OUT Packet
  if (gintsts & USB_GINTSTS_OEPINT) {
    msk = (((USB->DAINT & USB->DAINTMSK) >> 16) & 0xFFFF);
    num = 0;

    do {
      if ((msk >> num) & 1) {
        ep_int = USB_DOEPINT(num) & USB->DOEPMSK;
        ptr_ep = &ep[EP_ID(num)];

        // Setup phase done interrupt
        if (ep_int & USB_DOEP_INT_SETUP) {
          ptr_ep->dataSize = 0;
          USB_DOEPINT(num) = USB_DOEP_INT_SETUP;
          cbEndpointEvent(num, ARM_USBD_EVENT_SETUP);
        }

        // Transfer complete interrupt
        if (ep_int & USB_DOEP_INT_XFERCOMPL) {
          USB_DOEPINT(num) = USB_DOEP_INT_XFERCOMPL;
          if (USB_EP_OUT_TYPE(num) != ARM_USB_ENDPOINT_ISOCHRONOUS) {
            if (ptr_ep->dataSize != 0)
              USBD_EndpointReadSet(num);
            else {
              ptr_ep->flags &= ~USBD_EP_FLAG_BUSY;
              cbEndpointEvent(num, ARM_USBD_EVENT_OUT);
            }
          }
        }
      }
      num++;
    } while (msk >> num);
  }

  // IN Packet
  if (gintsts & USB_GINTSTS_IEPINT) {
    msk = (USB->DAINT & USB->DAINTMSK & 0xFFFF);
    num = 0;

    do {
      if ((msk >> num) & 1) {
        ep_int = USB_DIEPINT(num) & USB->DIEPMSK;
        ptr_ep = &ep[EP_ID(num | ARM_USB_ENDPOINT_DIRECTION_MASK)];

        // IN Endpoint NAK effective
        if (ep_int & USB_DIEP_INT_INEPNAKEFF) {
          if (ptr_ep->in_NAK) {
            ptr_ep->in_NAK = 0;

            val = 0;
            ptr_ep_in = &ep[0];
            for (i=0; i < USBD_EP_NUM; i++) {
              val |= ptr_ep_in->in_NAK;
              ptr_ep_in += 2;
            }

            // if no more forced NAKs, disable IN NAK effective interrupt
            if (!val) USB->DIEPMSK &= ~USB_DIEPMSK_INEPNAKEFFMSK;

            // If Data available, write Data
            if ( ptr_ep->dataSize || (!ptr_ep->buffer && !ptr_ep->dataSize)) {
              USBD_WriteToFifo(num | ARM_USB_ENDPOINT_DIRECTION_MASK);
            }
          }
          USB_DIEPINT(num) = USB_DIEP_INT_INEPNAKEFF;
        }

        // Transmit completed
        if (ep_int & USB_DIEP_INT_XFERCOMPL) {
          USB_DIEPINT(num) = USB_DIEP_INT_XFERCOMPL;
          if (ptr_ep->dataSize == 0) {
            ptr_ep->buffer =  NULL;
            ptr_ep->flags &= ~USBD_EP_FLAG_BUSY;
            cbEndpointEvent(num | ARM_USB_ENDPOINT_DIRECTION_MASK, ARM_USBD_EVENT_IN);
          } else {
            USBD_WriteToFifo(num | ARM_USB_ENDPOINT_DIRECTION_MASK);
          }
        }
      }
      num++;
    } while (msk >> num);
  }
}

ARM_DRIVER_USBD Driver_USBD0 = {
  USBD_GetVersion,
  USBD_GetCapabilities,
  USBD_Initialize,
  USBD_Uninitialize,
  USBD_PowerControl,
  USBD_DeviceConnect,
  USBD_DeviceDisconnect,
  USBD_DeviceGetState,
  USBD_DeviceRemoteWakeup,
  USBD_DeviceSetAddress,
  USBD_ReadSetupPacket,
  USBD_EndpointConfigure,
  USBD_EndpointUnconfigure,
  USBD_EndpointStall,
  USBD_EndpointTransfer,
  USBD_EndpointTransferGetResult,
  USBD_EndpointTransferAbort,
  USBD_GetFrameNumber
};
