/* -----------------------------------------------------------------------------
 * Copyright (c) 2013-2014 ARM Ltd.
 *
 * This software is provided 'as-is', without any express or implied warranty.
 * In no event will the authors be held liable for any damages arising from
 * the use of this software. Permission is granted to anyone to use this
 * software for any purpose, including commercial applications, and to alter
 * it and redistribute it freely, subject to the following restrictions:
 *
 * 1. The origin of this software must not be misrepresented; you must not
 *    claim that you wrote the original software. If you use this software in
 *    a product, an acknowledgment in the product documentation would be
 *    appreciated but is not required.
 *
 * 2. Altered source versions must be plainly marked as such, and must not be
 *    misrepresented as being the original software.
 *
 * 3. This notice may not be removed or altered from any source distribution.
 *
 *
 * $Date:        28. August 20144
 * $Revision:    V1.00
 *
 * Project:      USB common (Device and Host) module for SiliconLabs EFM32GGxxx
 * -------------------------------------------------------------------------- */

/* History:
 *  Version 1.00
 *    - Initial release
 */

#include "em_device.h"

#include "em_cmu.h"
#include "em_emu.h"
#include "em_gpio.h"

#include "Driver_USB.h"

#include "RTE_Device.h"
#include "RTE_Components.h"

static   bool     pins_configured_mask = 0;
volatile uint32_t USB_role = ARM_USB_ROLE_NONE;

__weak void USBH_IRQ (uint32_t gintsts) {};
__weak void USBD_IRQ (uint32_t gintsts) {};

/**
  \fn          void USB_IRQHandler (void)
  \brief       USB Interrupt Routine (IRQ).
*/
void USB_IRQHandler (void) {
  uint32_t gintsts;

  gintsts = USB->GINTSTS & USB->GINTMSK;

  switch (USB_role) {
    case ARM_USB_ROLE_HOST:
      USBH_IRQ (gintsts);
      break;
    case ARM_USB_ROLE_DEVICE:
      USBD_IRQ (gintsts);
      break;
    case ARM_USB_ROLE_NONE:
      break;
  }
}

/**
  \fn          bool USB_PinsConfigure (uint8_t pins_mask)
  \brief       Configure single or multiple USB Pin(s).
  \param[in]   Mask of pins to be configured (possible masking values:
               ARM_USB_PIN_DP, ARM_USB_PIN_DM, ARM_USB_PIN_VBUS, 
               ARM_USB_PIN_OC, ARM_USB_PIN_ID)
  \return      true = success, false = fail
*/
bool USB_PinsConfigure (uint8_t pins_mask) {

#if (RTE_USB)
#if (RTE_USB_HOST && !RTE_USB_VBUS_PIN)
#error   Configure VBUS Power On/Off pin for USB OTG Full-speed in RTE_Device.h!
#endif
#else
#error   Enable USB Full-speed in RTE_Device.h!
#endif

  if ((pins_configured_mask ^ pins_mask) & ARM_USB_PIN_DP) {
    // Init PHY. Enable USB pins (DP and DM)
    CMU->HFCORECLKEN0    |= CMU_HFCORECLKEN0_USB;
    USB->ROUTE           |= USB_ROUTE_PHYPEN;
    pins_configured_mask |= ARM_USB_PIN_DP | ARM_USB_PIN_DM;
  }
  if ((pins_configured_mask ^ pins_mask) & ARM_USB_PIN_DM) {
    // Init PHY. Enable USB pins (DP and DM)
    CMU->HFCORECLKEN0    |= CMU_HFCORECLKEN0_USB;
    USB->ROUTE           |= USB_ROUTE_PHYPEN;
    pins_configured_mask |= ARM_USB_PIN_DM | ARM_USB_PIN_DP;
  }
  if ((pins_configured_mask ^ pins_mask) & ARM_USB_PIN_ID) {
    pins_configured_mask |= ARM_USB_PIN_ID;
  }
  if ((pins_configured_mask ^ pins_mask) & ARM_USB_PIN_VBUS) {
    // Enable VBUS pin
    CMU_ClockEnable( cmuClock_GPIO, true );
    GPIO_PinModeSet( RTE_USB_VBUS_PORT, RTE_USB_VBUS_BIT, gpioModePushPull, 0);
    CMU->HFCORECLKEN0    |= CMU_HFCORECLKEN0_USB;
    USB->ROUTE           |= USB_ROUTE_VBUSENPEN;
    pins_configured_mask |= ARM_USB_PIN_VBUS;
  }
  if ((pins_configured_mask ^ pins_mask) & ARM_USB_PIN_OC) {
    GPIO_PinModeSet( RTE_USB_OC_PORT, RTE_USB_OC_BIT, gpioModeInput, 0 );
    pins_configured_mask |= ARM_USB_PIN_OC;
  }

  return true;
}

/**
  \fn          bool USB_PinsUnconfigure (uint8_t pins_mask)
  \brief       Unconfigure to reset settings single or multiple USB Pin(s).
  \param[in]   Mask of pins to be unconfigured (possible masking values:
               ARM_USB_PIN_DP, ARM_USB_PIN_DM, ARM_USB_PIN_VBUS, 
               ARM_USB_PIN_OC, ARM_USB_PIN_ID)
  \return      true = success, false = fail
*/
bool USB_PinsUnconfigure (uint8_t pins_mask) {

  if ((pins_configured_mask & pins_mask) & ARM_USB_PIN_DP) {
    // Disable USB pins (DP and DM)
    USB->ROUTE           &= ~USB_ROUTE_PHYPEN;
    pins_configured_mask &= ~(ARM_USB_PIN_DP | ARM_USB_PIN_DM);
  }
  if ((pins_configured_mask & pins_mask) & ARM_USB_PIN_DM) {
    // Disable USB pins (DP and DM)
    USB->ROUTE           &= ~USB_ROUTE_PHYPEN;
    pins_configured_mask &= ~(ARM_USB_PIN_DM | ARM_USB_PIN_DP);
  }
  if ((pins_configured_mask & pins_mask) & ARM_USB_PIN_ID) {
    pins_configured_mask &= ~ARM_USB_PIN_ID;
  }
  if ((pins_configured_mask & pins_mask) & ARM_USB_PIN_VBUS) {
    // Disable VBUS pin
    USB->ROUTE &= ~USB_ROUTE_VBUSENPEN;
    GPIO_PinModeSet( RTE_USB_VBUS_PORT, RTE_USB_VBUS_BIT, gpioModeDisabled, 0 );
    pins_configured_mask &= ~ARM_USB_PIN_VBUS;
  }
  if ((pins_configured_mask & pins_mask) & ARM_USB_PIN_OC) {
    /* Enable VBUS overcurrent flag pin. */
    GPIO_PinModeSet( RTE_USB_OC_PORT, RTE_USB_OC_BIT, gpioModeDisabled, 0 );
    pins_configured_mask &= ~ARM_USB_PIN_OC;
  }

  return true;
}

/**
  \fn          bool USB_PinVbusOnOff (bool state)
  \brief       Drive VBUS Pin On/Off.
  \param[in]   state    State On/Off (true = On, false = Off)
  \return      true = success, false = fail
*/

bool USB_PinVbusOnOff (bool state) {
#if (RTE_USB_VBUS_PIN)
  return true;
#else
  return false;
#endif
}
